--[[	Spydon Hex Viewer Module
	by SDPhantom
	http://www.phantomweb.org	]]
------------------------------------------

local Name,Addon=...;

----------------------------------
--[[	Unicode Pattern Data	]]
----------------------------------
local GetUnicodePositionData; do--	function(str,tbl)
	local UnicodePatterns={
		"[\192-\223]";--	0b110xxxxx +1 byte
		"[\224-\239]";--	0b1110xxxx +2 bytes
		"[\240-\247]";--	0b11110xxx +3 bytes (RFC 3629 stops here)
		"[\248-\251]";--	0b111110xx +4 bytes
		"[\252-\253]";--	0b1111110x +5 bytes
	};

--	Add additional payload bytes	0b10xxxxxx
	for i,j in ipairs(UnicodePatterns) do UnicodePatterns[i]=j..("[\128-\191]"):rep(i); end

	function GetUnicodePositionData(str,tbl)
		tbl=table.wipe(tbl or {});
		for i,j in ipairs(UnicodePatterns) do
			local offset=0;
			repeat
				local s,e,code=str:find("("..j..")",offset+1);
				if s then tbl[s],offset={e,code},e; end
			until not s;
		end
		return tbl;
	end
end

------------------------------------------
--[[	Grid Button Template Scripts	]]
------------------------------------------
do--	function SpydonHexViewerGridButton_OnEnter(self)
	local function ToBinary(val,pad)
		pad=pad or 0;
		local str="";
		repeat
			str=(val%2)..str;
			val=bit.rshift(val,1);
		until val<=0;
		return ("0"):rep(pad-#str)..str;
	end

	function SpydonHexViewerGridButton_OnEnter(self)
		local parent,id=self:GetParent(),self:GetID();
		self.LinkedButton:LockHighlight();
		parent.Highlight=self;

		parent.BytePositionLabel:SetFormattedText("%s Position",self.Type);
		parent.BytePositionText:SetFormattedText("%02X\n%d",id-1,id);
		parent.ByteConversionText:SetFormattedText("%02X\n%1$03d\n%s",self.Byte,ToBinary(self.Byte,8));
	end
end

function SpydonHexViewerGridButton_OnLeave(self)
	local parent=self:GetParent();
	self.LinkedButton:UnlockHighlight();
	if parent.Highlight==self then
		parent.Highlight=nil;

		parent.BytePositionLabel:SetText("Byte Position");
		parent.BytePositionText:SetText("");
		parent.ByteConversionText:SetText("");
	end
end

------------------------------------------
--[[	Main Frame (Created by Tab API)	]]
------------------------------------------
function SpydonHexViewerFrame_OnLoad(self)
	Addon.CreateTab("Hex Viewer",self);

--	Labels
	for i=0,15 do--			Column Labels (Always shown)
		local x,obj=i*25+45;
		obj=self:CreateFontString(nil,"OVERLAY","SpydonGridLabelFontNormal");
		obj:SetJustifyH("CENTER");
		obj:SetPoint("BOTTOM",self,"TOPLEFT",x,4);
		obj:SetFormattedText("%02X",i);

		obj=self:CreateTexture(nil,"ARTWORK");
		obj:SetTexture("Interface\\FriendsFrame\\WhoFrame-ColumnTabs");
		obj:SetTexCoord(0,5/64,0,19/32);
		obj:SetPoint("BOTTOMLEFT",self,"TOPLEFT",x-14,4);
		obj:SetSize(5,19);

		obj=self:CreateTexture(nil,"ARTWORK");
		obj:SetTexture("Interface\\FriendsFrame\\WhoFrame-ColumnTabs");
		obj:SetTexCoord(5/64,58/64,0,19/32);
		obj:SetPoint("BOTTOMLEFT",self,"TOPLEFT",x-9,4);
		obj:SetSize(16,19);

		obj=self:CreateTexture(nil,"ARTWORK");
		obj:SetTexture("Interface\\FriendsFrame\\WhoFrame-ColumnTabs");
		obj:SetTexCoord(58/64,62/64,0,19/32);
		obj:SetPoint("BOTTOMLEFT",self,"TOPLEFT",x+7,4);
		obj:SetSize(4,19);
	end

	do	local txt;--		Prefix Row (Always shown)
		txt=self:CreateFontString(nil,"OVERLAY","SpydonGridLabelFontNormal");
		txt:SetPoint("TOPLEFT");
		txt:SetJustifyH("LEFT");
		txt:SetText("Prefix");
	end

	self.Rows={}; for i=0,15 do--	Payload Rows (Show/Hide based on payload length)
		local txt=self:CreateFontString(nil,"OVERLAY","SpydonGridLabelFontNormal");
		txt:SetPoint("TOPRIGHT",self,"TOPLEFT",30,(i+1)*-18);
		txt:SetJustifyH("RIGHT");
		txt:SetFormattedText("0x%X0",i);

		self.Rows[i+1]=txt;
	end

--	Grids
	self.PrefixGrid={}; for i=1,16 do--	Prefix Grid
		local btn={
			CreateFrame("Button",nil,self,"SpydonHexViewerHexGridButtonTemplate",i);
			CreateFrame("Button",nil,self,"SpydonHexViewerCharGridButtonTemplate",i);
		};

		btn[1]:SetPoint("TOP",self,"TOPLEFT",(i-1)*25+45,2);
		btn[2]:SetPoint("TOP",self,"TOPLEFT",(i-1)*12+445,2);

		btn[1].Type="Prefix";	btn[1]:SetFormattedText("%02X",i-1);	btn[1].LinkedButton=btn[2];
		btn[2].Type="Prefix";	btn[2]:SetFormattedText("%X",i-1);	btn[2].LinkedButton=btn[1];

		self.PrefixGrid[i]=btn;
	end

	self.PayloadGrid={}; for i=1,255 do--	Payload Grid
		local btn={
			CreateFrame("Button",nil,self,"SpydonHexViewerHexGridButtonTemplate",i);
			CreateFrame("Button",nil,self,"SpydonHexViewerCharGridButtonTemplate",i);
		};

		btn[1]:SetPoint("TOP",self,"TOPLEFT",((i-1)%16)*25+45,math.floor((i-1)/16)*-18-16);
		btn[2]:SetPoint("TOP",self,"TOPLEFT",((i-1)%16)*12+445,math.floor((i-1)/16)*-18-16);

		btn[1].Type="Payload";	btn[1]:SetFormattedText("%02X",i-1);	btn[1].LinkedButton=btn[2];
		btn[2].Type="Payload";	btn[2]:SetFormattedText("%X",(i-1)%16);	btn[2].LinkedButton=btn[1];

		self.PayloadGrid[i]=btn;
	end

--	Update Helper Functions
	local UpdateGrid; do--	function(grid,str)
		local UnicodeTable={};
		function UpdateGrid(grid,str)
			local len,unicodeend=#str,0;
			GetUnicodePositionData(str,UnicodeTable);
			for i,j in ipairs(grid) do
				if i<=len then
					j[1]:Show();		j[2]:Show();

					local byte=str:byte(i);
					if UnicodeTable[i] then
						local dat=UnicodeTable[i]; unicodeend=dat[1];
						j[1]:SetNormalFontObject("SpydonHexFontUnicode");
						j[2]:SetNormalFontObject("SpydonCharFontUnicode");
						j[2]:SetText(dat[2]);
					elseif i<=unicodeend then
						j[1]:SetNormalFontObject("SpydonHexFontUnicode");
						j[2]:SetNormalFontObject("SpydonCharFontUnicode");
						j[2]:SetText("");
					elseif byte>=32 and byte<127 then
						local char=string.char(byte)
						j[1]:SetNormalFontObject("SpydonHexFontNormal");
						j[2]:SetNormalFontObject("SpydonCharFontNormal");
						j[2]:SetText(char);
					else
						j[1]:SetNormalFontObject("SpydonHexFontNonPrint");
						j[2]:SetNormalFontObject("SpydonCharFontNonPrint");
						j[2]:SetText("?");
					end
					j[1]:SetFormattedText("%02X",byte);
					j[1].Byte,j[2].Byte=byte,byte;

					if j[1]:IsMouseOver() then
						SpydonHexViewerGridButton_OnEnter(j[1]);
					elseif j[2]:IsMouseOver() then
						SpydonHexViewerGridButton_OnEnter(j[2]);
					end
				else
					j[1]:UnlockHighlight();	j[2]:UnlockHighlight();
					j[1]:Hide();		j[2]:Hide();
				end
			end
		end
	end

--	Update Script
	Addon.RegisterEvent(--	Current Message Update
		function(func,event,...)
			local id,ptr=Addon.GetCurrentMessage();
			if ptr then
				self.ErrorText:Hide();

				for i,j in ipairs(self.Rows) do j:SetShown((i-1)*16<=#ptr[5]); end
				UpdateGrid(self.PrefixGrid,ptr[4]);
				UpdateGrid(self.PayloadGrid,ptr[5]);

				if self.Highlight and not self.Highlight:IsShown() then
					SpydonHexViewerGridButton_OnLeave(self.Highlight);
				end
			else
				self.ErrorText:Show();
				for i,j in ipairs(self.Rows) do j:Show(); end
				for i,j in ipairs(self.PrefixGrid) do j[1]:Hide(); j[2]:Hide(); end
				for i,j in ipairs(self.PayloadGrid) do j[1]:Hide(); j[2]:Hide(); end
			end
		end
		,"SPYDON_MESSAGELIST_UPDATE"
		,"SPYDON_MESSAGELIST_SELECTED"
		,"SPYDON_MESSAGELIST_HIGHLIGHTED"
	);
end
