--[[	Spydon Core
	by SDPhantom
	http://www.phantomweb.org	]]
------------------------------------------

local Name,Addon=...;
local Title=GetAddOnMetadata(Name,"Title");
Spydon=Addon;

--	Setup SavedVariable
Addon.Options={};
Spydon_Options=Addon.Options;

--------------------------
--[[	Local Funcs	]]
--------------------------
local FormatDisplayString; do--	function(str)
--	Converts non-prints to red "?"
	local function helper(str) return ("|cffff0000%s|r"):format(("?"):rep(#str)); end
	function FormatDisplayString(str) return str:gsub("[^\32-\126]+",helper); end
end

----------------------------------
--[[	Main Frame Scripts	]]
----------------------------------
function SpydonFrame_OnLoad(self)
--	Apply final fixes to XML objects
	self.Title:SetText(Title);
	SetPortraitToTexture(self.Portrait,self.Portrait:GetTexture());

--	Register as a UI Panel
	UIPanelWindows["SpydonFrame"]={area="left"};
	Addon.MainFrame=self;

--	Slash Command (Frame doesn't exist until now)
	function SlashCmdList.SPYDON()
		_G[self:IsShown() and "HideUIPanel" or "ShowUIPanel"](self);
	end
	SLASH_SPYDON1="/spydon";
	SLASH_SPYDON2="/spy";

--	Tab API
	self.Tabs={}; self.CurrentTab=1;
	function self.CreateTab(name,frame)
--		If we don't have a frame, create one from template
		frame=frame or CreateFrame("Frame",nil,self,"SpydonSubFrameTemplate");

--		Tab Creation
		local id=#self.Tabs+1;
		local tab=CreateFrame("Button",nil,self,"SpydonTabButtonTemplate",id);
		if id>1 then
			tab:SetPoint("TOPLEFT",self.Tabs[id-1],"TOPRIGHT",-8,0);
			frame:Hide();
		else
			tab:SetPoint("TOPLEFT",self,"BOTTOMLEFT",15,12);
			tab:Disable();
			frame:Show();
		end

--		Store frame, set name, resize
		tab.Frame=frame;
		tab:SetText(name);
		tab:SetWidth(tab:GetFontString():GetWidth()+40);

--		Store tab and return frame in case we had to create one
		self.Tabs[id]=tab;
		return frame;
	end

	function self.SelectTab(id)
--		Enable old, disable new, swap frames
		local new=self.Tabs[id];
		if id~=self.CurrentTab and new then
			local old=self.Tabs[self.CurrentTab];
			old:Enable();	old.Frame:Hide();
			new:Disable();	new.Frame:Show();
			self.CurrentTab=id;
		end
	end
	Addon.CreateTab=self.CreateTab;
	Addon.SelectTab=self.SelectTab;

--	Other Frame/API Functions
	function self.GetCurrentMessage()
--		This is where the order of these IDs are taken into consideration (Default to last message received)
		local id=self.HighlightedID or self.SelectedID or #Addon.MessageCache;
		return id,Addon.MessageCache[id];
	end
	Addon.GetCurrentMessage=self.GetCurrentMessage;

--	Create Buttons
	self.Buttons={}; for i=1,15 do
		local btn=CreateFrame("Button",nil,self,"SpydonMessageButtonTemplate",i);
		btn:SetPoint("TOPLEFT",23,(i-1)*-20-105);
		self.Buttons[i]=btn;
	end

--[[	Message structure is as follows:
	1:	Timestamp
	2:	Type (IN, OUT, ECHO)
	3:	Stacktrace (IN/ECHO = Prefix registration stack, OUT = Stack sending message)
	4:	Prefix
	5:	Message
	6:	Channel
	7:	IN/ECHO: Sender		OUT: Target (WHISPER channel only)
--]]

--	Update Functions
	Addon.RegisterEvent(--	Autoscroll
		function(func,event,...)
			local val,off,min,max=#Addon.MessageCache-#self.Buttons,self.Scroll:GetValue(),self.Scroll:GetMinMaxValues();
			if val>max then
--				Update max and enable scrolling
				self.Scroll:SetMinMaxValues(0,val);
				self.Scroll:Enable();

--				Autoscroll or just validate (Updates Up/Down buttons)
				if off>=max and not self.SelectedID then
					self.Scroll:SetValue(val);
				else
					self.Scroll:Validate(off);
				end
			end
		end
		,"SPYDON_MESSAGELIST_UPDATE"
	);
	Addon.RegisterEvent(--	Message List Update
		function(func,event,...)
			local len,sel,off=#Addon.MessageCache,self.SelectedID,self.Scroll:GetValue();
			if event~="SPYDON_MESSAGELIST_UPDATE" or len<=#self.Buttons then
--				If we have more messages than buttons, this will start updating from autoscroll
				for i,j in ipairs(self.Buttons) do
					local id=off+i;
					local ptr=Addon.MessageCache[id];
					if ptr then
--						Color button background if there's something wrong
						local tex=j:GetNormalTexture();
						if Addon.PrefixCallStack[ptr[4]] then			tex:SetVertexColor(1,1,1);--	OUT messages will have stack data
						elseif IsAddonMessagePrefixRegistered(ptr[4]) then	tex:SetVertexColor(1,1,0);--	Only can happen if something registered a prefix before we loaded
						else							tex:SetVertexColor(1,0,0);
						end

--						Set text and colors
						j.Prefix:SetFontObject(Addon.ChannelFonts[ptr[6]] or SpydonLabelFontNormalSmall);
						j.Prefix:SetText(FormatDisplayString(ptr[4]));
						j.Player:SetFontObject(Addon.DirectionFonts[ptr[2]] or SpydonLabelFontNormalSmall);
						j.Player:SetText(ptr[7]:match("^[^%-]*"));

--						Update selection and show
						j[id==sel and "LockHighlight" or "UnlockHighlight"](j);
						j:Show();

--						If scrolled and mouse is still highlighting button, update with new highlight ID
						if j:IsMouseOver() then
							self.HighlightedID=id;
							Addon.FireEvent("SPYDON_MESSAGELIST_HIGHLIGHTED",id);
						end
					else
--						Nothing to show
						j:UnlockHighlight();
						j:Hide();
					end
				end
			end

--			We need this updating even if the buttons don't need to be
			self.MessageCount:SetFormattedText("%d-%d/%d",math.min(off+1,len),math.min(off+#self.Buttons,len),len);
		end
		,"SPYDON_PREFIX_ADDED"
		,"SPYDON_MESSAGELIST_UPDATE"
		,"SPYDON_MESSAGELIST_SCROLL"
	);
	Addon.RegisterEvent(--	Current Message Update
		function(func,event,...)
			local id,ptr=self.GetCurrentMessage();
			if ptr then
--				Show error if there is any
				if not IsAddonMessagePrefixRegistered(ptr[4]) then
					self.ErrorText:SetFontObject(SpydonWarningFont);
					self.ErrorText:SetFormattedText("Warning: |cffffcc00%s|r not registered!",FormatDisplayString(ptr[4]));
				elseif not Addon.PrefixCallStack[ptr[4]] then
					self.ErrorText:SetFontObject(SpydonErrorFont);
					self.ErrorText:SetFormattedText("Error: |cffffcc00%s|r call stack not found!",FormatDisplayString(ptr[4]));
				else
					self.ErrorText:SetText("");
				end

--				Message header text and color
				self.MessageHeader:SetFontObject(Addon.ChannelFonts[ptr[6]] or SpydonLabelFontNormalSmall);
				self.MessageHeader:SetFormattedText("<%s #%d> [%s][%s]: %s<%s>|r%s"
					,date("%H:%M:%S",ptr[1]),id
					,ptr[6],ptr[7]
					,Addon.GetFontColorCode(Addon.DirectionFonts[ptr[2]] or SpydonLabelFontNormalSmall)
					,ptr[2],FormatDisplayString(ptr[4])
				);

--				Message prefix/payload sizes
				self.MessagePrefixLength:SetFormattedText("Prefix:%d",#ptr[4]);
				self.MessagePayloadLength:SetFormattedText("Payload:%d",#ptr[5]);
			else
--				No message, show with error text too
				self.ErrorText:SetText("No Messages Found!");
				self.MessageHeader:SetText("");
				self.MessagePrefixLength:SetText("");
				self.MessagePayloadLength:SetText("");
			end
		end
		,"SPYDON_MESSAGELIST_UPDATE"
		,"SPYDON_MESSAGELIST_SELECTED"
		,"SPYDON_MESSAGELIST_HIGHLIGHTED"
	);
end

function SpydonFrame_OnShow(self)
	local selid,offset,min,max=self.SelectedID,self.Scroll:GetValue(),self.Scroll:GetMinMaxValues();

--	Unselect our old button (ignore if offscreen)
	local btn=self.Buttons[(selid or 0)-offset];
	if btn then btn:UnlockHighlight(); end

--	Select/Highlight nothing
	self.SelectedID=nil;
	self.HighlightedID=nil;

--	Show first tab, try to set autoscroll, and fire selection update
	self.SelectTab(1);
	self.Scroll:SetValue(max);
	Addon.FireEvent("SPYDON_MESSAGELIST_SELECTED",nil,selid);
end

do--	function SpydonFrame_OnUpdate(self)
--	This entire thing is just for the memory usage display
	local MemLabels={"Bytes","KB","MB","GB","TB"};
	local function GetMemoryLabel(val)--	Base is in KB with multipliers of 1024
		for i=#MemLabels,1,-1 do
--			Calculate new value for this unit
			local newval=val/1024^(i-2);

--			Threshold is 1000/1024, the ratio of difference between metric and binary unit steps
			if newval>=0.9765625 then return newval,MemLabels[i]; end
		end

--		Shouldn't get here, but default to "0 Bytes"
		return 0,MemLabels[1];
	end

	local Timer=0;
	function SpydonFrame_OnUpdate(self,elapsed)
--		This should be faster than comparing GetTime()'s
		Timer=Timer-elapsed;

		if Timer<=0 then
--			Profiler needs to gather info before GetAddOnMemoryUsage() can be used
			UpdateAddOnMemoryUsage();
			self.MemoryText:SetFormattedText("%04.2f %s",GetMemoryLabel(GetAddOnMemoryUsage(Name)));

--			Reset timer to 10 seconds
			Timer=10;
		end
	end
end

------------------------------------------
--[[	Tab Button Template Scripts	]]
------------------------------------------
--	Wrapper for the SelectTab API
function SpydonTabButton_OnClick(self) self:GetParent().SelectTab(self:GetID()); end

------------------------------------------
--[[	Message Button Template Scripts	]]
------------------------------------------
function SpydonMessageButton_OnClick(self)
--	Base data
	local parent,btnid=self:GetParent(),self:GetID();
	local selid,offset=parent.SelectedID,parent.Scroll:GetValue();
	local msgid=offset+btnid;

--	Unselect our last button (ignore if offscreen)
	local last=parent.Buttons[(selid or 0)-offset];
	if last then last:UnlockHighlight(); end

	if selid==msgid then
--		Select nothing and fire event
		parent.SelectedID=nil;
		Addon.FireEvent("SPYDON_MESSAGELIST_SELECTED",nil,selid);
	else
--		Select self and fire event
		self:LockHighlight();
		parent.SelectedID=msgid;
		Addon.FireEvent("SPYDON_MESSAGELIST_SELECTED",msgid,selid);
	end
end

function SpydonMessageButton_OnEnter(self)
	local parent=self:GetParent();
	local id=parent.Scroll:GetValue()+self:GetID();

--	Update self-highlight, fire highlight update event
	parent.HighlightedID=id;
	Addon.FireEvent("SPYDON_MESSAGELIST_HIGHLIGHTED",id);
end

function SpydonMessageButton_OnLeave(self)
	local parent=self:GetParent();
	if parent.HighlightedID==parent.Scroll:GetValue()+self:GetID() then
--		Clear highlight and fire event
		parent.HighlightedID=nil;
		Addon.FireEvent("SPYDON_MESSAGELIST_HIGHLIGHTED",nil);
	end
end

----------------------------------
--[[	Hyperlink Registration	]]
----------------------------------
do	local oldfunc=SetItemRef;
	function SetItemRef(...)--	(link,text,button,frame)
		local tab,id=(...):match("^Spydon:(%d+):(%d+)$");
		if tab and id then
--			Convert to numbers and show our frame.
			tab,id=tonumber(tab),tonumber(id);
			local frame=Addon.MainFrame;
			ShowUIPanel(frame);

--			Handle the tab ID
			if tab>0 then frame.SelectTab(tab); end

--			Handle the message ID
			if id>0 then
				local selid,offset=frame.SelectedID,frame.Scroll:GetValue();

--				Find the old selected button and unlock the highlight (ignore if scrolled offscreen)
				local selbtn=frame.Buttons[(selid or 0)-offset];
				if selbtn then selbtn:UnlockHighlight(); end

				if id==selid then
--					Unselect and fire the update event
					frame.SelectedID=nil;
					Addon.FireEvent("SPYDON_MESSAGELIST_SELECTED",nil,selid);
				else
--					Select this button
					frame.SelectedID=id;

--					Try to scroll to our new selected button
					frame.Scroll:SetValue(id-14);
					offset=frame.Scroll:GetValue();

--					The scroll update event may have already taken care of this, but we need to make sure
					selbtn=frame.Buttons[id-offset];
					if selbtn then selbtn:LockHighlight(); end

--					Fire the selection update event
					Addon.FireEvent("SPYDON_MESSAGELIST_SELECTED",id,selid);
				end
			end
		else
--			Pass to the old function
			return oldfunc(...);
		end
	end
end

--------------------------
--[[	Initialization	]]
--------------------------
Addon.RegisterEvent(
	function(func,event,...)
		if (...)==Name then
--			Sync with loaded options
			Addon.Options=Spydon_Options;

--			Welcome message (With new hyperlink)
			print(("|cffff8800|HSpydon:0:0|h[%s]|h|r Loaded! Type /spy to open."):format(Title));

--			Fire first update
			Addon.FireEvent("SPYDON_MESSAGELIST_UPDATE");
			Addon.UnregisterEvent(func,event);
		end
	end
	,"ADDON_LOADED"
);
